package org.greenstone.gsdl3.gs3build.indexers;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import org.w3c.dom.*;

import org.greenstone.mg.*;

import org.greenstone.gsdl3.gs3build.doctypes.DocumentID;
import org.greenstone.gsdl3.gs3build.doctypes.DocumentInterface;
import org.greenstone.gsdl3.gs3build.doctypes.HTMLDocument;
import org.greenstone.gsdl3.gs3build.doctypes.METSDocument;
import org.greenstone.gsdl3.gs3build.metadata.*;
import org.greenstone.gsdl3.gs3build.xpointer.XPointer;
import org.greenstone.gsdl3.gs3build.util.DOMUtils;
import org.greenstone.gsdl3.util.GSXML;
import org.greenstone.gsdl3.util.Misc;
import org.greenstone.gsdl3.util.Processing;

// for debug
import org.greenstone.gsdl3.util.XMLConverter;

public class MGIndexer //extends AbstractIndexer
    implements IndexerInterface
{ 
    int          pass;
    int          documentSeqNo;
    int          sectionSeqNo;
    boolean      firstDocument;
    String       outputDirectory;
    StringBuffer indexBuffer;
    File         textDirectory;
    File         indexDirectory;
    String       indexStem;
    String       textStem;
    List         indexes;
    String       overallName; 
    
    MGIndex current_index = null;
    MGPassesWrapper mgPasses;
    
    static final char END_OF_DOCUMENT = (char) 2;
    //static final char END_OF_SECTION  = (char) 3; // actually this is end of para for mg - don't use this yet
    static final char END_OF_STREAM   = (char) 4;

    public static final String MG_INDEX_TYPE = "mg";
    public static final String INDEX_FILE_STEM = "index";
    
    class MGIndex
    { 
	String name=null;
	String level=null;
	List fields=null;
	boolean error = false;// assume built until we get an error
	
	public MGIndex(Element index_element) {
	    
	    this.fields = new ArrayList();
	    this.name = index_element.getAttribute(GSXML.NAME_ATT);
	    if (this.name.equals("")) {
		// TODO make this dynamic
		this.name = "xx";
	    }
	    NodeList children = index_element.getChildNodes();
	    for (int c = 0; c < children.getLength(); c ++) {
		Node child = children.item(c);
		
		if (child.getNodeType() == Node.ELEMENT_NODE) {
		    String name = child.getNodeName();
	  
		    if (name.equals(GSXML.LEVEL_ELEM)) {
			this.level = DOMUtils.getNodeChildText(children.item(c));
		    }
		    else if (name.equals(GSXML.FIELD_ELEM)) {
			String fieldName = DOMUtils.getNodeChildText(children.item(c));
			this.fields.add(fieldName);
		    }
		}
	    }
	}

	public MGIndex(String name, String level, List fields)
	{ 
	    this.name  = name;
	    this.level = level;
	    this.fields = fields;
	}
	
	// old gs2 style config - can we get rid of this??
	public MGIndex(String indexLabel)
	{ 
	    int colonAt = indexLabel.indexOf(':');
	    
	    if (colonAt < 0) {
		System.err.println("MGIndex(): invalid index specification: "+indexLabel);
		return;
	    }
	    String field_string = indexLabel.substring(colonAt+1);
	    String [] field_list = field_string.split(",");
	    this.fields = new ArrayList();
	    for (int i=0; i<field_list.length; i++) {
		this.fields.add(field_list[i]);
	    }
	    this.level = indexLabel.substring(0, colonAt);
	    //createIndexName
	    StringBuffer new_name = new StringBuffer();
	    new_name.append(Character.toLowerCase((char) this.level.charAt(0)));
	    int c, w;
	    w = 0;
	    c = 0;
	    while (c < field_string.length() && w < 2) {
		char ch = field_string.charAt(c);
		
		ch = Character.toLowerCase(ch);
		if (Character.isLetter(ch)) {
		    if (ch != 'a' && ch != 'e' && ch != 'i' &&
			ch != 'o' && ch != 'u') {
			new_name.append(ch);
			w++;
		    }
		}
		c ++;
	    }
	    this.name = new_name.toString();
	}

	
	public String getLevel()
	{ 
	    return this.level;
	}

	public List getFields()
	{ 
	    return this.fields;
	}

	public String getName()
	{ 
// 	    if (this.name==null || this.name.equals("")) {
// 		createIndexName();
// 	    }
	    return this.name;
	}

	public boolean hasError() {
	    return this.error;
	}
	public void setError(boolean b) {
	    this.error = b;
	}

// 	private void createIndexName() {
// 	    StringBuffer new_name = new StringBuffer();
// 	    new_name.append(Character.toLowerCase((char) this.level.charAt(0)));

// 	    int c, w;
// 	    w = 0;
// 	    c = 0;
// 	    String [] fields_concat = this.fields.toArray
// 	    while (c < this.field.length() && w < 2) {
// 		char ch = this.field.charAt(c);

// 		ch = Character.toLowerCase(ch);
// 		if (Character.isLetter(ch)) {
// 		    if (ch != 'a' && ch != 'e' && ch != 'i' &&
// 			ch != 'o' && ch != 'u') {
// 			new_name.append(ch);
// 			w++;
// 		    }
// 		}
// 		c ++;
// 	    }
// 	    this.name = new_name.toString();
// 	}
 

    } // MGIndex

    public MGIndexer(String name)
    { 
	this.indexes = new ArrayList();
	this.overallName = name; 
    }

    public String getIndexType()
    { 
	return MG_INDEX_TYPE;
    } 

    public String getName()
    { 
	return this.overallName; 
    }
    
    public boolean configure(Node search_node)
    {
	NodeList index_children = GSXML.getChildrenByTagName(search_node, GSXML.INDEX_ELEM);
	
	// add a text 'index' - we should be able to turn this off in the config file? actually mg needs a text index
	ArrayList list = new ArrayList();
	list.add("text");
	MGIndex index = new MGIndex("text", "section", list);
	indexes.add(index);
	for (int i = 0; i < index_children.getLength(); i ++) {
	    Element index_elem = (Element)index_children.item(i);
	    index = new MGIndex(index_elem);
	    if (index.getName() != null && index.getLevel() != null && index.getFields()!= null) {
		
		indexes.add(index);
	    } else {
		System.err.println("invalid index spec, not including"+new XMLConverter().getPrettyString(index_elem));
	    }
	}
	// TODO make sure all index names are unique
	return true;
    }

    /**
     *  The output directory should be (collection)/building/text/ for
     *  normal Greenstone builds.
     *
     *  @param <code>String</code> the label to configure
     *  @param <code>String</code> the value...
     */ 
    public boolean configure(String label, String value)
    { 
	if (label.equals(IndexerManager.outputDir)) {
	    this.outputDirectory = value;
	    this.pass = 0;
    
	    // attempt to ensure that the text subdirectory exists
	    this.textDirectory = new File(outputDirectory, "text");
	    if (!textDirectory.exists()) {
		if (!textDirectory.mkdir()) {
		    return false;
		}
	    }
	    else if (!textDirectory.isDirectory()) {
		return false;
	    }
	    this.textStem = this.textDirectory.getPath() + File.separator + INDEX_FILE_STEM;

	    // Sign to the user which mg directory is being used...
	    System.out.println("Output MG directory is " + this.textStem);
	}
	else if (label.equals(IndexerInterface.GS2_INDEX_LABEL)) {
	    this.indexes.add(new MGIndex(value));
	}

	return true;
    }


    private Node recurseDOM(DocumentInterface metsDoc, Node node, 
			    AbstractStructure structure, StringBuffer textBuffer,
			    StringBuffer extraBuffer, String namespace) 
	//String name, String namespace, String field)
    {
	List fields = current_index.getFields();
	// send out the ctrl-c...if this is
	if (structure.getStructureType().equals(METSDivision.DIVISION_TYPE)) {
	    // try doing this for all index types
	    // actually we should only need to do this once ????
	    if (this.pass == 0) {
	    //if ((this.currentIndexName != null)) { // && this.level != null && this.level.equals(IndexerInterface.SECTION_LEVEL)) { //name.startsWith("s")) {
		METSDivision division = (METSDivision) structure;

		// get the division metadata block
		METSDescriptive descriptive;
		String metadataId = division.getDefaultMetadataReference();
		if (metadataId == null) {
		    descriptive = metsDoc.getDocumentMetadata().createDescriptive(division.getLabel());
		    division.addMetadataReference(descriptive.getID());
		}
		else {
		    // Get the descriptive item...
		    descriptive = metsDoc.getDocumentMetadata().getDescriptiveById(metadataId);
		}
	    
		descriptive.addMetadata("gsdl3", "mgseqno", this.overallName + "." + Integer.toString(this.sectionSeqNo));
	    
		metsDoc.setChanged(true);
		//metsDoc.setModified(true);
		//	System.out.println("Assigning " + this.sectionSeqNo + " to " + metsDoc.getID() + " " + division.getLabel());
	    } // section level
	
	    // append an 'end of section' marker
	    //textBuffer.append(END_OF_SECTION);
	    this.sectionSeqNo ++;
	
	    // for document-level indexes, always append an 'end of document' tag at the
	    // end of the document for each section.  Otherwise, each section is followed
	    // by an end of document character.  This ensures that all indexes use the
	    // same document numbering...
	    if (this.current_index.getLevel().equals(IndexerInterface.DOCUMENT_LEVEL)) {
		extraBuffer.append(END_OF_DOCUMENT);
	    }
	    else {
		textBuffer.append(END_OF_DOCUMENT);
		this.documentSeqNo ++;
	    }
	    
	    // produce the body here for metadata output of divisions - in the case of
	    // text output, that will happen below...
	    
	    if (fields.size()>1 || !((String)fields.get(0)).equals("text")) {
		// if there is only text, don't do this
		METSDescriptive descriptive;
	
		METSDivision division = (METSDivision) structure;

		String metadataId = division.getDefaultMetadataReference();
		// are there other metadata refs to get??
		descriptive = metsDoc.getDocumentMetadata().getDescriptiveById(metadataId);
		if (descriptive != null) {
		    for (int i=0; i<fields.size(); i++) {
			String field = (String)fields.get(i);
			if (field.equals("text")) {
			    continue;
			}
			List values = descriptive.getMetadata(namespace, field);
			if (values != null) {	
			    Iterator valueIter = values.iterator();
			    while (valueIter.hasNext()) {
				String value = valueIter.next().toString();
				textBuffer.append(value);
				textBuffer.append(" ");
			    }
			}
		    }
		}
	    }
	}
	
	// go through our children as required...
	Iterator children = structure.getChildIterator();
	Node startNode;
	boolean index_text = fields.contains("text");
	while (children.hasNext()) {
	    AbstractStructure child = (AbstractStructure) children.next();
      
	    // get xpointer for child 
	    // get start position node
	    if (metsDoc.getDocumentType() == "METS"){
		startNode = ((METSDocument) metsDoc).getSectionStartNode((METSDivision) child);	  
	    } else {
		startNode = ((HTMLDocument) metsDoc).getSectionStartNode((METSDivision) child);
	    }
	    
	    // while this node isn't the child's start node, produce the 
	    // HTML node text, if in text field mode...
	    if (index_text) {
		while (node != startNode) {
		    XPointer.printNode(node, textBuffer, false);
		    node = XPointer.getNextNode(node);
		}
	    }
	    
	    // recurse to child
	    node = this.recurseDOM(metsDoc, node, child, textBuffer, extraBuffer, namespace); // name, namespace, field);
	} // while next child

	// close a document - the actual closing \B will be done by the main
	// loop, so only a required \C is printed here...
	// why have we got STRUCTURE_TYPE here and DIVISION_TYPE above????
	if (structure.getStructureType().equals(METSStructure.STRUCTURE_TYPE)) {
	    if (index_text) {
		while (node != null) {
		    XPointer.printNode(node, textBuffer, false);
		    node = XPointer.getNextNode(node);
		}
	    }
      
	    //textBuffer.append(END_OF_SECTION);
	    this.sectionSeqNo ++;
      
	}
	return node;
    }

    private String prepareDOM(DocumentInterface metsDoc, Document document, METSStructure structure, String namespace)
    {
	//  String name, String namespace, String field)
	StringBuffer extraBuffer = new StringBuffer();
	Node node = document.getDocumentElement();
	StringBuffer textBuffer = new StringBuffer();
	
	this.recurseDOM(metsDoc, node, structure, textBuffer, extraBuffer, namespace); //name, namespace, field);
	textBuffer.append(extraBuffer.toString());
	return textBuffer.toString();
    }

    /**
     *  Index a single document; the document interface can be used to extract individual
     *  metadata items etc. as required or desired and index those instead or as well as
     *  the body text of the document.
     */
    public boolean indexDocument(DocumentID docID, DocumentInterface document)
    {
	
	if (!this.firstDocument) {
	    this.indexBuffer.append(END_OF_DOCUMENT);
	    mgPasses.processDocument(indexBuffer.toString());
	    this.indexBuffer.delete(0, this.indexBuffer.length());
	    
	}

	String docText = null;
	// set the mgseqno if first pass
	if (this.pass == 0) {
	    document.removeAllMetadata("gsdl3", "mgseqno");
	    document.addDocumentMetadata("gsdl3", "mgseqno", this.overallName+"."+Integer.toString(this.sectionSeqNo));
	}

	this.sectionSeqNo ++;
	
	//long start = System.currentTimeMillis();
	Document domDocument = document.getDOMDocument();
	if (domDocument != null) {
	    System.err.println("dom doc is not null");
	    METSStructure sections = document.getDocumentStructure().getStructure("Section");
	    if (sections != null) {
		System.err.println("sections are not null");
		docText = this.prepareDOM(document, domDocument, sections, "gsdl3"); //this.name, "gsdl3", this.field);
		//	System.out.println(docText);
	    }
	}
	//long finish = System.currentTimeMillis();
	//System.err.println("dom doc = "+ Long.toString(finish-start));
	//start = System.currentTimeMillis();
	if (docText == null) {
	    System.err.println("dom doc or sections was null - asking for doc text");
	    StringBuffer doc_text_buffer = new StringBuffer();
	    List fields = this.current_index.getFields();
	    for (int i=0; i<fields.size(); i++) {
		String field = (String)fields.get(i);
		if (field.equals("text")) {
		    doc_text_buffer.append(document.getDocumentText());
		}  else {
		    // its a metadata - do namespace properly!!
		    List values = document.getDocumentMetadataItem("gsdl3", field);
		    if (values != null) {
			Iterator valueIter = values.iterator();
			while (valueIter.hasNext()) {
			    String value = valueIter.next().toString();
			    doc_text_buffer.append(value);
			}
		    }
		}
	    } // for each field
	    docText = doc_text_buffer.toString();
	    sectionSeqNo ++;
	}
	//finish = System.currentTimeMillis();
	//System.err.println("whole doc = "+ Long.toString(finish-start));
	
	this.indexBuffer.append(docText);
	// remember that we're not on the first document, 
	this.firstDocument = false;
	this.documentSeqNo ++;

	return true;
    }

    /**
     *  Initialise the pass: open required files, check status
     */
    public boolean startPass(int passNumber) {
    	
	
	this.pass = passNumber;
	this.firstDocument = true;
	this.documentSeqNo = 1;
	this.sectionSeqNo  = 1;
	
	this.mgPasses = new MGPassesWrapper();
	this.indexBuffer = new StringBuffer();
	int indexNo = this.pass/2;
	this.current_index = null;
	
	this.current_index = (MGIndex) this.indexes.get(indexNo);
	if (this.current_index.hasError()) {
	    // an error has already occurred for this index, don't continue
	    System.out.println("pass "+this.pass+": aborted due to errors in the previous pass");
	    return false;
	}
	// attempt to ensure that the text/index subdirectory exists
	this.indexDirectory = new File(outputDirectory, current_index.getName());
	if (!indexDirectory.exists()) {
	    if (!indexDirectory.mkdir()) {
		return false;
	    }
	}
	else if (!indexDirectory.isDirectory()) {
	    return false;
	}

	this.indexStem = this.indexDirectory.getPath() + File.separatorChar + INDEX_FILE_STEM;  // TODO: modify for index
	if (this.pass == 0) {
	    // first pass, also set up the textStem
	    this.textDirectory = this.indexDirectory;
	    this.textStem = this.indexStem;
	}
	mgPasses.setFileName(this.indexStem);
	if (!Misc.isWindows()) {
	    mgPasses.setBasePath("/");
	}
	int mgPass = this.pass < 2 ? this.pass : ((this.pass % 2) + 2);
	
	mgPasses.setBufferSize(100000);
 	
	switch (mgPass) {
	case 0:
	    // -b 100000 -T1
	    mgPasses.addPass(MGPassesWrapper.TEXT_PASS_1);
	    
	    
	    break;
	    
	case 1:
	    // -b 100000 -T2
	    mgPasses.addPass(MGPassesWrapper.TEXT_PASS_2);
	    break;
	    
	case 2:
	    // -b 100000 -2 -m 32 -s 0 -G -t 10 -N1
	    mgPasses.addPass(MGPassesWrapper.INDEX_PASS_1);
	    mgPasses.setInvfLevel(MGPassesWrapper.INVF_LEVEL_2);
	    mgPasses.setStemOptions(MGPassesWrapper.STEMMER_ENGLISH, MGPassesWrapper.NO_STEM_OR_CASE);
	    mgPasses.setInversionMemLimit(32);
	    mgPasses.ignoreSGMLTags(true);
	    break;
	    
	case 3:
	    // -b 100000 -2 -c 3 -G -t 10 -N2
	    mgPasses.addPass(MGPassesWrapper.INDEX_PASS_2);
	    mgPasses.setInvfLevel(MGPassesWrapper.INVF_LEVEL_2);
	    mgPasses.ignoreSGMLTags(true);
	    break;
	}
	
	mgPasses.init();
	System.out.println("Starting Pass " + this.pass);
	return true;
    }
  
    /**
     *  Complete a pass - reset file counters, close files, etc.
     */
    public boolean endPass(int passNumber) {
	try {
	    this.indexBuffer.append(END_OF_DOCUMENT);
	    mgPasses.processDocument(indexBuffer.toString());
	    this.indexBuffer.delete(0, this.indexBuffer.length());
	    Thread.sleep(1000); // what for??
	}
	catch (InterruptedException ex) {
	    System.out.println(ex);
	}
	mgPasses.finish();
	try {
	    Thread.sleep(1000);
	} catch (Exception e) {}
	
	int exit_value = 0;
	System.out.println("Pass " + this.pass + " completed with " + exit_value);
	if (exit_value !=0) {
	    //assume something has gone wrong, don't continue
	    current_index.setError(true);
	    return false;
	}
	int mgPass = this.pass < 2 ? this.pass : ((this.pass % 2) + 2);
	String osextra = "";
	if (!Misc.isWindows()) {
	    osextra = " -d / ";
	}
	
	switch (mgPass) {
	    
	case 0:
	    System.out.println("Compressing dictionary");
	    exit_value = Processing.runProcess("mg_compression_dict -f " + this.textStem + osextra + " -S -H -2 -k 5120");
	    if (exit_value == 0) {
		System.out.println("Compressed dictionary successfully written");
	    } else {
		System.err.println("Error from mg_compression_dict: " + exit_value);
		current_index.setError(true);
		
		return false;
	    }
	    break;
	    
	case 2:
	    System.out.println("Creating perfect hash");
	    exit_value = Processing.runProcess("mg_perf_hash_build -f " + this.indexStem+osextra);
	    if (exit_value ==0) {
		System.out.println("Perfect hashes completed");
	    } else {
		System.err.println("Unable to build the perfect hash");
		current_index.setError(true);
		return false;
	    }
	    break;
	    
	case 3:
	    System.out.println("Writing weights file");
	    exit_value = Processing.runProcess("mg_weights_build -f " + this.indexStem + " -t " + this.textStem + osextra);
	    if (exit_value ==0) {
		System.out.println("Weights file successfully written");
	    } else {
		System.err.println("Unable to create weights file");
		current_index.setError(true);
		return false;
	    }
	    
	    System.out.println("Creating inverted dictionary");
	    exit_value = Processing.runProcess("mg_invf_dict -f " + this.indexStem + osextra);
	    if (exit_value ==0) {
		System.out.println("Inverted dictionary file successfully written");
	    } else {
		System.out.println("Unable to create inverted dictionary file");
		current_index.setError(true);
		return false;
	    }
	    
	    System.out.println("Creating Stem indexes");
	    exit_value = Processing.runProcess("mg_stem_idx -b 4096 -s1 -f " + this.indexStem+osextra);
	    if (exit_value == 0) {
		System.out.println("Stemmed index 1 successfully written");
	    } else {
		System.out.println("Unable to create stemmed index 1");
		current_index.setError(true);
		return false;
	    }
	    
	    exit_value = Processing.runProcess("mg_stem_idx -b 4096 -s2 -f " + this.indexStem+osextra);
	    if (exit_value == 0) {
		System.out.println("Stemmed index 2 successfully written");
	    } else {
		System.out.println("Unable to create stemmed index 2");
		current_index.setError(true);
		return false;
	    }
	    exit_value = Processing.runProcess("mg_stem_idx -b 4096 -s3 -f " + this.indexStem+osextra);
	    if (exit_value == 0) {
		System.out.println("Stemmed index 3 successfully written");
	    } else {
		System.out.println("Unable to create stemmed index 3");
		current_index.setError(true);
		return false;
	    }
	    
	    break;
	} // switch

	mgPasses = null;
	return true;
    }
    
    /**
     *  Do any tidying up
     */
    public void tidyup()
    {
    }
    
    /**
     *  Return the number of passes required for this index.
     */
    public int getNumberOfPasses()
    { 
	//return 2 + this.indexes.size() * 2;
	return this.indexes.size()*2;
    }

    public boolean addServiceDescriptions(org.w3c.dom.Element service_rack_list)
    {
	Document doc = service_rack_list.getOwnerDocument();

	// generate the list of indexes
	Element index_list = doc.createElement(GSXML.INDEX_ELEM+GSXML.LIST_MODIFIER);
	boolean found_index = false;
	String def_index = ""; // the default index will just be the first one created for now.
	for (int i=1; i<this.indexes.size(); i++) {
	    // start at 1: 0 will be the text index
	    MGIndex index = (MGIndex)this.indexes.get(i);
	    if (!index.hasError()) {
		Element e = doc.createElement(GSXML.INDEX_ELEM);
		e.setAttribute(GSXML.NAME_ATT, index.getName());
		index_list.appendChild(e);
		if (found_index == false) {
		    // this is the first index
		    found_index = true;
		    def_index = index.getName();
		}
	    }
	}
	
	if (!found_index) {
	    // no indexes were able to be created, so we can't use them or the text
	    return false;
	}
	Element default_index = doc.createElement("defaultIndex");
	default_index.setAttribute(GSXML.NAME_ATT, def_index);
	Element base_index_name = doc.createElement("baseIndexPrefix");
	base_index_name.setAttribute(GSXML.NAME_ATT, overallName);
	Element index_stem = doc.createElement("indexStem");
	index_stem.setAttribute(GSXML.NAME_ATT, INDEX_FILE_STEM);
	
	Element search_service_elem = doc.createElement(GSXML.SERVICE_CLASS_ELEM);
	Element retrieve_service_elem = doc.createElement(GSXML.SERVICE_CLASS_ELEM);
	service_rack_list.appendChild(search_service_elem);
	service_rack_list.appendChild(retrieve_service_elem);

	search_service_elem.setAttribute(GSXML.NAME_ATT, "GS3MGSearch");  
	search_service_elem.appendChild(index_list);
	search_service_elem.appendChild(default_index);
	search_service_elem.appendChild(base_index_name);
	search_service_elem.appendChild(index_stem);
	
	retrieve_service_elem.setAttribute(GSXML.NAME_ATT, "GS3MGRetrieve");
	retrieve_service_elem.appendChild(default_index.cloneNode(true));
	retrieve_service_elem.appendChild(base_index_name.cloneNode(true));
	retrieve_service_elem.appendChild(index_stem.cloneNode(true));

	return true;
    }
    
}
    
